use std::time::{SystemTime};
use rand::Rng;
use std::fs::File;
use std::path::Path;
use std::fs;

#[macro_use]
extern crate serde_derive;

extern crate rand;
extern crate serde;
extern crate serde_json;


/*
UNITS
    mass = kilogram
    time = 1 day
    distance = meter
    speed = meter / day
*/

// Time constants
const SECOND: f32 = MINUTE/60.0;
const MINUTE: f32 = HOUR/60.0;
const HOUR: f32 = DAY/24.0;
const DAY: f32 = 1.0;
const MONTH: f32 = DAY*30.0;
const YEAR: f32 = MONTH*12.0;
// weight constants
const KG: f32 = 1.0;
const TON: f32 = KG*1000.0;




#[derive(Serialize, Deserialize, Debug)]
struct Config {
    biome: Biome,
    lifeforms: Vec<Lifeform>,
    simtime: f32,
    simsteps: f32
}


// Diet types a lifeform could have
#[derive(Serialize, Deserialize, Debug, Clone)]
enum Diets {
    Carnivore,
    Herbivore,
    Omnivore,
    Photovore
}

// The types of lifeforms that may exist
#[derive(Serialize, Deserialize, Debug, Clone)]
enum LifeformKind {
    Fauna,
    Flora,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
struct Rect {
    x: f32,
    y: f32,
    width: f32,
    height: f32
}




/*
Biome - a container for lifeforms.
Ideally, I wanted to have next to no logic here, but I realized my lifeforms needed access to the biome,
for example to re-deposit their weight back on death, to query for food, etc. I'd like to pass this to a 
lifeform so it can access it and modify it.
*/
// Maybe biome needs a hashmap that can be locked to hold lifeforms
#[derive(Serialize, Deserialize, Debug, Clone)]
struct Biome {
    water: f32,
    biomass: f32,
    lifeforms: Vec<Lifeform>,
    rect: Rect
}

    

impl Biome {
    fn tick(&mut self, timescale: f32){


        for mut lifeform in &mut self.lifeforms {
            if lifeform.alive {
                if self.water > 0.0 {
                    self.water -= lifeform.water_intake * timescale;
                }

                if lifeform.digestive_content / lifeform.digestive_capacity < 0.2 {
                    // hungry, let's eat
                    //this is expensive to calc...

                lifeform.mass += lifeform.nutrient_intake * timescale;
                }


                if lifeform.mass <= lifeform.mass_starve {
                    lifeform.alive = false;
                    println!("Died of starvation: {:?}", lifeform);
                }

                // age
                lifeform.age += 1.0 * timescale;
                if lifeform.age > lifeform.maxage {
                    lifeform.alive = false;
                    println!("Died of age: {:?}", lifeform);
                }



                move_random(&mut lifeform, &self.rect, timescale);
            } else {
                // fuck, we're dead
                self.biomass += lifeform.mass;
                lifeform.mass = 0.0;
            }

    }
    
    }

}

#[derive(Serialize, Deserialize, Debug, Clone)]
struct Lifeform {
    name: String,           //just for fun
    mass: f32,              //how much it weights in kg
    mass_starve: f32,       //at this mass, starvation begins
    digestive_capacity: f32,//a food "buffer"; stomach or tree trunk
    digestive_content: f32, //how much the digestive buffer can store
    age: f32,               //age in days
    maxage: f32,            //maximum age in days
    water_intake: f32,      //how much water it consumes per tick
    nutrient_intake: f32,   //how much food it consumes per tick
    diet: Diets,            //the type of stuff it eats
    kind: LifeformKind,     //the kind of lifeform it is
    alive: bool,            //whether it lives or not
    position: Vec<f32>,     //where it hangs out
    speed: f32,             //how fast it can move (or not),
    ocurrence_per_sq_km: f32//distribution
}


// TODO: add position
impl Lifeform {

    /*
    Here is what the lifeform 'does' per tick. It consumes food, gains weight...
    i would like to be able to pass the biome to the lifeform as a reference to that
    it could interact with it.
    */
    // fn tick(&mut self, biome: &mut Biome){

    // }
}

impl Default for Lifeform {


    fn default() -> Lifeform {
        Lifeform {
            name: "Generic Plant".to_string(),
            mass: 1.0,
            mass_starve: 0.0,
            digestive_capacity: 0.0,
            digestive_content: 0.0,
            age: 0.0,
            water_intake: 1.0/DAY as f32,
            nutrient_intake: 1.0/DAY as f32,
            diet: Diets::Herbivore,
            kind: LifeformKind::Flora,
            maxage: YEAR,
            alive: true,
            position: vec![0.0, 0.0, 0.0],
            speed: 0.0,
            ocurrence_per_sq_km: 1.0
        }
    }
}


fn move_random(lf: &mut Lifeform, rect: &Rect, timescale: f32) {
        let mut rng = rand::thread_rng();
        let rx = rng.gen_range::<f32>(-1.0, 1.0);
        let rz = rng.gen_range::<f32>(-1.0, 1.0);
        lf.position[0] += (rx * lf.speed) * timescale as f32;
        lf.position[2] += (rz * lf.speed) * timescale as f32;
        // where is clamp
        if lf.position[0] > rect.width+rect.x {lf.position[0] = rect.width+rect.x}; // x > right bound
        if lf.position[2] > rect.height+rect.y {lf.position[2] = rect.height+rect.y}; // z > upper bound
        if lf.position[0] < rect.x {lf.position[0] = rect.x}; // x < left bound
        if lf.position[2] < rect.y {lf.position[2] = rect.y}; // z < lower bound

    }



fn sim() {

    // add some trees
    let mut rng_tree = rand::thread_rng();
    let tree = Lifeform {
        name: "Tree".to_string(),
        age: rng_tree.gen_range::<f32>(1.0*MONTH, 10.0*YEAR),
        maxage: rng_tree.gen_range::<f32>(5.0*YEAR, 50.0*YEAR),
        mass: rng_tree.gen_range::<f32>(50.0, 6.0*TON),
        water_intake: rng_tree.gen_range::<f32>(50.0, 150.0)/DAY,
        diet: Diets::Photovore,
        alive: true,
        kind: LifeformKind::Flora,
        ..Default::default()
        };


    // add deer
    let mut rng_deer = rand::thread_rng();
    let deer = Lifeform {
        name: "Deer".to_string(),
        kind: LifeformKind::Fauna,
        age: rng_deer.gen_range::<f32>(1.0*MONTH, 5.0*YEAR),
        maxage: rng_deer.gen_range::<f32>(2.0*YEAR, 6.0*YEAR),
        mass: rng_deer.gen_range::<f32>(4.0, 17.0),
        water_intake: rng_deer.gen_range::<f32>(0.5, 1.0)/DAY,
        speed: rng_deer.gen_range::<f32>(0.0, 10000.0/HOUR),
        digestive_capacity: 2.0,
        mass_starve: 5.0,
        ocurrence_per_sq_km: 15.0,
        ..Default::default()};


    let gen_config = false;

    if gen_config {
        let config_rect = Rect {x: 0.0, y: 0.0, width: 100.0, height: 100.0};
    
        let config_biome = Biome {water: 1000.0, lifeforms: Vec::new(), biomass: TON, rect: config_rect};
        let mut config_lifeforms = Vec::new();
        config_lifeforms.push(deer);
        config_lifeforms.push(tree);

        let config = Config {biome: config_biome, lifeforms: config_lifeforms, simtime: 1000.0, simsteps: 1.0};
        let serialized_config = serde_json::to_string_pretty(&config).unwrap();
        
        fs::write("config.json", serialized_config).expect("Unable to write config file");
    }
    
    
    //biome.lifeforms.push(deer);
    //biome.lifeforms.push(tree);

    // let deserialized: Plant = serde_json::from_str(&serialized).unwrap();
    //println!("deserialized = {:?}", deserialized);


    // Load config file
    let json_config_path = Path::new("config.json");
    let config_file = File::open(json_config_path).expect("config file not found");
    let deserialized_config: Config = serde_json::from_reader(config_file).expect("error while reading json config");
    // println!("{:?}", deserialized_config);


    let mut loaded_biome = deserialized_config.biome;

    for lf in deserialized_config.lifeforms {

        let sq_km = (loaded_biome.rect.width*loaded_biome.rect.width)/1000000.0;
        let num_instances: i32 = (sq_km * lf.ocurrence_per_sq_km).floor() as i32;
        println!("Instantiating from {:?} {:?} times", lf.name, num_instances);
        for _i in 0..num_instances {
            let mut instance = lf.clone();
            loaded_biome.lifeforms.push(instance);
        }
    

    }

    //biome.lifeforms.push(tree);
    println!("running {:?} days for {:?} lifeforms", deserialized_config.simtime * deserialized_config.simsteps, loaded_biome.lifeforms.len());
    for _i in 1..deserialized_config.simtime as i32 {
        loaded_biome.tick(deserialized_config.simsteps);
    }

    let serialized = serde_json::to_string(&loaded_biome).unwrap();
    //println!("serialized = {}", serialized);
    fs::write("sim.json", serialized).expect("Unable to write file");
}


fn main() {
    //let start = SystemTime::now();
    println!("\n\n=== === === Starting simulation === === ===");
    sim();
}
